package net.sf.sql2java.data.node;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.sql2java.data.MapPointer;
import net.sf.sql2java.data.Path;
import net.sf.sql2java.data.Pointer;

@SuppressWarnings("unchecked")
public class BasicNodelet
	extends HashMap<String, Map<String, Nodelet>> 
	implements Nodelet 
{
	private static final long serialVersionUID = 4315960428210027147L;
	private String name;
	private String type;
	
	private Nodelet parent;
	
// -- Constructors ------------------------------------------------------------
	
	public BasicNodelet(String type, String name) {
		super();

		this.type = type;
		this.name = name;
	}
	
// -- Conversion --------------------------------------------------------------
	
	@Override
	public String toString() {
		StringBuffer result = new StringBuffer();

		if (parent != null)
			result.append(parent.toString()).append(Path.SEPARATOR);
		
		result.append(String.format(MapPointer.FORMAT, type, name));
		
		return result.toString();
	}
	
// -- Map Manipulation --------------------------------------------------------
	
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Nodelet> T get(Path path) {
		List<Nodelet> list = list(path);
		
		if (list.size() > 1)
			throw new RuntimeException(
					String.format("Multiple nodes returned for path %s", path.toString()));
		
		return (T) list.get(0);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Nodelet> List<T> list(Path path) {
		// -- create the result object
		List<T> result = new LinkedList<T>();
		
		// -- get the pointer we are interested in
		Pointer pointer = path.getElement(0);
		Path newPath = path.getSubPath(1);
		
		// -- get the nodelet corresponding to the nodelet's name
		Map<String, Nodelet> nodes = get(pointer.getName());
		
		// -- return if no nodes could be found
		if (nodes == null) return result;
		
		// -- once we have the pointer, we'll need to check the pointer type.
		// -- depending on the pointer type, the nodelet has to be a collection
		// -- nodelet or a basic nodelet
		if (pointer instanceof MapPointer) {
			// -- we have to retrieve a specific node inside the map
			Nodelet node = nodes.get(((MapPointer) pointer).getId());
			
			if (path.getElements().size() > 1)
				result.addAll((List<T>) node.list(newPath));
			else
				result.add((T) node);
			
		} else {
			// -- if we get a regular node, we'll have to return all the values
			// -- for the given node
			if (path.getElements().size() > 1) {
				// -- we need to add the nodes of all the possible paths to the
				// -- result map.
				for (Nodelet node : nodes.values())
					result.addAll((Collection<T>) node.list(newPath));
			} else
				result.addAll((Collection<T>) nodes.values());
		}
		
		// -- return the result
		return result;
	}
	
	@Override
	public void put(Path path, Nodelet node) {
		// -- get the pointer we are interested in
		Pointer pointer = path.getElement(0);
		Path newPath = path.getSubPath(1);
		
		// -- get the nodelet corresponding to the nodelet's name
		Map<String, Nodelet> nodes = get(pointer.getName());
		
		// -- return if no nodes could be found
		if (nodes == null) 
			nodes = new HashMap<String, Nodelet>();
		
		// -- we are at the end. Insert the node here
		if (newPath.getElements().size() == 0) {
			// -- set the node's parent
			node.setParent(this);
			
			// -- insert the node
			nodes.put(node.getName(), node);
		
			// -- persist the changes
			put(pointer.getName(), nodes);
			
		// -- let's keep on going down the tree
		} else {
			// -- the pointer has to be a map pointer, because we can't add a
			// -- pointer to a collection of nodes
			if (! (pointer instanceof MapPointer))
				throw new RuntimeException("Pointer has to be a mappointer to go down the tree");
			
			MapPointer mp = (MapPointer) pointer;
			
			Nodelet childNode = nodes.get(mp.getId());
			
			if (childNode == null)
				throw new RuntimeException(
						String.format("No node found for path %s", path.toString()));
			
			childNode.put(newPath, node);
		}
	}
	
	public void put(Nodelet node) {
		String strPath = String.format("%s(%s)", node.getType(), node.getName());
		
		put(Path.createPath(strPath), node);
	}
	
	public void remove(Path path) {
		// -- get the pointer we are interested in
		Pointer pointer = path.getElement(0);
		Path newPath = path.getSubPath(1);
		
		// -- the pointer has to be a map pointer, because we can't remove a
		// -- pointer to a collection of nodes
		if (! (pointer instanceof MapPointer))
			throw new RuntimeException("Pointer has to be a mappointer to go down the tree");
		
		MapPointer mp = (MapPointer) pointer;
		
		// -- get the nodelet corresponding to the nodelet's name
		Map<String, Nodelet> nodes = get(pointer.getName());
		
		// -- return if no nodes could be found
		if (nodes == null) 
			nodes = new HashMap<String, Nodelet>();
		
		// -- we are at the end. Insert the node here
		if (newPath.getElements().size() == 0) {
			// -- insert the node
			nodes.remove(mp.getId());
		
			// -- persist the changes
			put(mp.getName(), nodes);
			
		// -- let's keep on going down the tree
		} else {
			Nodelet childNode = nodes.get(mp.getId());
			
			if (childNode == null)
				throw new RuntimeException(
						String.format("No node found for path %s", path.toString()));
			
			childNode.remove(newPath);
		}
	}
	
	public boolean exists(Path path) {
		// -- get the pointer we are interested in
		Pointer pointer = path.getElement(0);
		Path newPath = path.getSubPath(1);
		
		// -- the pointer has to be a map pointer, because we can't remove a
		// -- pointer to a collection of nodes
		if (! (pointer instanceof MapPointer))
			throw new RuntimeException("Pointer has to be a mappointer to go down the tree");
		
		MapPointer mp = (MapPointer) pointer;
		
		// -- get the nodelet corresponding to the nodelet's name
		Map<String, Nodelet> nodes = get(pointer.getName());
		
		// -- return if no nodes could be found
		if (nodes == null) return false;
		
		// -- we are at the end. Insert the node here
		if (newPath.getElements().size() == 0) {
			return nodes.containsKey(mp.getId());
			
		// -- let's keep on going down the tree
		} else {
			Nodelet childNode = nodes.get(mp.getId());
			
			if (childNode == null)
				throw new RuntimeException(
						String.format("No node found for path %s", path.toString()));
			
			return childNode.exists(newPath);
		}
	}

	@Override
	public String getName() {
		return name;
	}
	
	@Override
	public String getType() {
		return this.type;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Nodelet> T getParent() {
		if (parent == null) return null;
		else return (T) parent;
	}

	@Override
	public void setParent(Nodelet parent) {
		this.parent = parent;
	}
}
